// GPL
// Author:  Yotam Medini  yotam.medini@gmail.com -- Created: 2006/March/10
#include "Alps.h"
#include <unistd.h>
#include <fcntl.h>
#include <cstring>
#include <errno.h>
#include <iostream>
#include <iomanip>

// Since xf86 headers do not respect C++ we:
#define bool     _bOol
#define class    _klAss
#define new      _nEw
#define private  priv8
#define public   publik
extern "C"
{
#define NEED_XF86_TYPES
#   include <linux/input.h>
#   include <xf86.h>
#   include <xf86Xinput.h>
#   include <xf86_OSproc.h>
#   include <mipointer.h>
#   include <xisb.h>
#   include <os.h>
#if defined(DEBUG)
#   include <xf86Optrec.h>
#endif
}
#undef bool
#undef class
#undef new
#undef private
#undef public

#include "../config.h"
#include "ButtonHandler.h"
#include "PadHandler.h"
#include "SharedMemConfig.h"
#include "conf.h"
#include "dbg.h"

using namespace std;

enum
{
    bitsPerLong = sizeof(long) * 8,
    nULongs     = (KEY_MAX + bitsPerLong-1) / bitsPerLong,
};


enum
es
{
// acknowledge for commands and parameter
    PS2_ACK                  = 0xFA,
    PS2_ERROR                = 0xFC,

// standard PS/2 commands
    PS2_CMD_RESET            = 0xFF,
    PS2_CMD_RESEND           = 0xFE,
    PS2_CMD_SET_DEFAULT      = 0xF6,
    PS2_CMD_DISABLE          = 0xF5,
    PS2_CMD_ENABLE           = 0xF4,
    PS2_CMD_SET_SAMPLE_RATE  = 0xF3,
    PS2_CMD_READ_DEVICE_TYPE = 0xF2,
    PS2_CMD_SET_REMOTE_MODE  = 0xF0,
    PS2_CMD_SET_WRAP_MODE    = 0xEE,
    PS2_CMD_RESET_WRAP_MODE  = 0xEC,
    PS2_CMD_READ_DATA        = 0xEB,
    PS2_CMD_SET_STREAM_MODE  = 0xEA,
    PS2_CMD_STATUS_REQUEST   = 0xE9,
    PS2_CMD_SET_RESOLUTION   = 0xE8,
    PS2_CMD_SET_SCALING_2_1  = 0xE7,
    PS2_CMD_SET_SCALING_1_1  = 0xE6
};


////////////////////////////////////////////////////////////////////////
class Config
{
 public:
    class Dim
    {
     public:
        Dim(): min(0), max(1), nc(1), dc(1), ne(1), de(1) {}
            unsigned    min;
            unsigned    max;
            unsigned    nc;
            unsigned    dc;
            unsigned    ne;
            unsigned    de;
            unsigned    jump;
    }; // Dim
    Config() : delay_ms(0) {}
    void      setOptions(pointer options);
    void      shm_set(shm_config::Handler&) const;
    void      shm_get(const shm_config::Handler&);
    unsigned  delay_ms;
    Dim       dim[2];
 private:
    unsigned  dimOption(pointer o, unsigned di, const char* name,
                        unsigned defval);
    unsigned  setIntOption(pointer o, const char* name, unsigned defval);
}; // Config


////////////////////////////////////////////////////////////////////////
void Config::setOptions(pointer o)
{
    using namespace confdefault;
    delay_ms = setIntOption(o, "EmulateMidButtonTime", EmulateMidButtonTime);
    for (unsigned  di = 0;  di != 2;  ++di)
    {
        static unsigned  min_xy_defval[2] = {xPadMin, yPadMin};
        static unsigned  max_xy_defval[2] = {xPadMax, yPadMax};
        static unsigned  xy_jump[2] = {xJump, yJump};
        Dim&  d = dim[di];
        d.min  = dimOption(o, di, "PadMin", min_xy_defval[di]);
        d.max  = dimOption(o, di, "PadMax", max_xy_defval[di]);
        d.nc   = dimOption(o, di, "CenterNumerator", 1);
        d.dc   = dimOption(o, di, "CenterDenominator", 1);
        d.ne   = dimOption(o, di, "EdgeNumerator", 1);
        d.de   = dimOption(o, di, "EdgeDenominator", 1);
        d.jump = dimOption(o, di, "Jump", xy_jump[di]);
    }
} // Config::setOptions


#if !defined(NO_SHM)
////////////////////////////////////////////////////////////////////////
void Config::shm_set(shm_config::Handler& shm) const
{
    using namespace shm_config;
    Handler::MailBox& mb = *shm.mailBox();

    mb.options[SHM_OPT_EmulateMidButtonTime] = delay_ms;
    mb.options[SHM_OPT_xPadMin]              = dim[0].min;
    mb.options[SHM_OPT_xPadMax]              = dim[0].max;
    mb.options[SHM_OPT_xCenterNumerator]     = dim[0].nc;
    mb.options[SHM_OPT_xCenterDenominator]   = dim[0].dc;
    mb.options[SHM_OPT_xEdgeNumerator]       = dim[0].ne;
    mb.options[SHM_OPT_xEdgeDenominator]     = dim[0].de;
    mb.options[SHM_OPT_xJump]                = dim[0].jump;
    mb.options[SHM_OPT_yPadMin]              = dim[1].min;
    mb.options[SHM_OPT_yPadMax]              = dim[1].max;
    mb.options[SHM_OPT_yCenterNumerator]     = dim[1].nc;
    mb.options[SHM_OPT_yCenterDenominator]   = dim[1].dc;
    mb.options[SHM_OPT_yEdgeNumerator]       = dim[1].ne;
    mb.options[SHM_OPT_yEdgeDenominator]     = dim[1].de;
    mb.options[SHM_OPT_yJump]                = dim[1].jump;
} // Config::shm_set


////////////////////////////////////////////////////////////////////////
void Config::shm_get(const shm_config::Handler& shm)
{
    using namespace shm_config;
    const Handler::MailBox& mb = *shm.mailBox();

    delay_ms    = mb.options[SHM_OPT_EmulateMidButtonTime];
    dim[0].min  = mb.options[SHM_OPT_xPadMin];
    dim[0].max  = mb.options[SHM_OPT_xPadMax];
    dim[0].nc   = mb.options[SHM_OPT_xCenterNumerator];
    dim[0].dc   = mb.options[SHM_OPT_xCenterDenominator];
    dim[0].ne   = mb.options[SHM_OPT_xEdgeNumerator];
    dim[0].de   = mb.options[SHM_OPT_xEdgeDenominator];
    dim[0].jump = mb.options[SHM_OPT_xJump];
    dim[1].min  = mb.options[SHM_OPT_yPadMin];
    dim[1].max  = mb.options[SHM_OPT_yPadMax];
    dim[1].nc   = mb.options[SHM_OPT_yCenterNumerator];
    dim[1].dc   = mb.options[SHM_OPT_yCenterDenominator];
    dim[1].ne   = mb.options[SHM_OPT_yEdgeNumerator];
    dim[1].de   = mb.options[SHM_OPT_yEdgeDenominator];
    dim[1].jump = mb.options[SHM_OPT_yJump];
} // Config::shm_get
#endif

////////////////////////////////////////////////////////////////////////
unsigned
Config::dimOption(pointer o, unsigned di, const char* name, unsigned val)
{
    char  xyname[0x20], *pxyname = &xyname[0];
    xyname[0] = "xy"[di];
    strcpy(pxyname + 1, name);
    val = setIntOption(o, pxyname, val);
    ADBG(pxyname << " -> " << val);
    return val;
} // Config::dimOption


////////////////////////////////////////////////////////////////////////
unsigned Config::setIntOption(pointer o, const char* name, unsigned defval)
{
    unsigned val = xf86SetIntOption(o, name, defval);
    xf86Msg(X_INFO, "alps: set %s := %d [default = %d]\n", name, val, defval);
    return val;
} // Config::setIntOption


////////////////////////////////////////////////////////////////////////
const char*  Alps::version = PACKAGE_VERSION;


////////////////////////////////////////////////////////////////////////
const int  Alps::_bhe2xbi[3] =
{
    1,  // ButtonHandler::ButtonLeft
    3,  // ButtonHandler::ButtonRight
    2   // ButtonHandler::ButtonMid
};

////////////////////////////////////////////////////////////////////////
void*  Alps::setup(pointer module, pointer options, int* errmaj, int* errmin)
{
    static InputDriverRec alpsIDC =
    {
        1,
        (char*)"alps",
        &identify,  // Identify
        &preinit,
        &uninit,
        0,  // module
        0   // refCount
    };


    ADBG("ALPS::setup module=" << module);
    xf86AddInputDriver(&alpsIDC, module, 0);
    return module;
} // Alps::setup


////////////////////////////////////////////////////////////////////////
//   read lines from   /proc/bus/input/devices
void Alps::identify(int flags)
{
    ADBG("identify flags=0x" << hex << flags << dec);
} // Alps::identify


////////////////////////////////////////////////////////////////////////
InputInfoPtr  Alps::preinit(InputDriverPtr drv, IDevPtr dev, int flags)
{
    ADBG("preinit flags=0x" << hex << flags << dec);
    InputInfoPtr  ii = 0;
    Alps*         alps = new Alps(drv, dev, flags);
    if (alps->ok())
    {
        ii = alps->inpinfo();
    }
    else
    {
        delete alps;
    }
    ADBG("ii=" << ii << ", alps=" << alps);
    return ii;
} // Alps::preinit


////////////////////////////////////////////////////////////////////////
void Alps::uninit(InputDriverPtr drv, InputInfoPtr inpInfo, int flags)
{
    Alps*  alps = thisAlps(inpInfo);
    ADBG("uninit alps=" << alps << ", flags=0x" << hex << flags << dec);
    delete alps;
} // Alps::uninit


////////////////////////////////////////////////////////////////////////
Alps*  Alps::thisAlps(DeviceIntPtr dev)
{
    inpinfo_t*  ii = (inpinfo_t*)(dev->publik.devicePrivate);
    return thisAlps(ii);
} // Alps::thisAlps


////////////////////////////////////////////////////////////////////////
Alps*  Alps::thisAlps(inpinfo_t* inpInfo)
{
    Alps*  p = (Alps*)(inpInfo->priv8);
    return p;
} // Alps::thisAlps


////////////////////////////////////////////////////////////////////////
Alps::Alps(driver_t* drv, IDevPtr dev, int flags) :
    _inpInfo(0),
    _error(0),
    _xisBuffer(0),
    _buttonHandler(0),
    _timer(0),
    _timer2BeCalled(0),
    _padHandler(0),
    _midEmulationDelayMilliSec(0),
    _shm(0)
{
    ADBG("Alps::Alps");
#if !defined(NO_SHM)
    _shm = shm_config::Handler::create();
#endif
    _timerCallParam[0] = TimerCallParam(this, 0);
    _timerCallParam[1] = TimerCallParam(this, 1);
    _inpInfo = xf86AllocateInput(drv, 0);
    if (_inpInfo == 0)
    {
        ErrorF("alps driver: xf86AllocateInput failed\n");
        setError(eAllocInp);
    }
    if (ok())
    {
        setInpInfo(dev);
        xf86Msg(X_INFO, "alps touchpad driver version %s\n", version);
        xf86CollectInputOptions(_inpInfo, 0, 0);
        xf86OptionListReport(_inpInfo->options);
#if defined(DEBUG)
        for (XF86OptionPtr opts = (XF86OptionPtr)(_inpInfo->options);
             opts;  opts = xf86nextOption(opts))
        {
            const char*  v = xf86optionValue(opts);
            ADBG("option=" << xf86optionName(opts) <<
                 ", v=" << (v ? v : "Null"));
        }
#endif
        openSerial(); // open the touchpad device
    }
    if (ok() && devReset() && newBuffer(0x100))
    {
        Config  conf;
        xf86ProcessCommonOptions(_inpInfo, _inpInfo->options);
        conf.setOptions(_inpInfo->options);
        _inpInfo->flags |= XI86_CONFIGURED;
        configure(conf);
#if !defined(NO_SHM)
        if (_shm->ok()) { conf.shm_set(*_shm); }
#endif
        ADBG("flags=0x" << hex << _inpInfo->flags << dec);
    }
    release();
} // Alps::Alps


////////////////////////////////////////////////////////////////////////
Alps::~Alps()
{
    ADBG("Alps::~Alps");
    release();
    delete _buttonHandler;
    TimerFree(_timer);
    delete _padHandler;
#if !defined(NO_SHM)
    delete _shm;
#endif
} // Alps::~Alps


////////////////////////////////////////////////////////////////////////
void Alps::release()
{
    ADBG("Alps::release");
    delBuffer();
    closeSerial();
} // Alps::release


////////////////////////////////////////////////////////////////////////
void Alps::setInpInfo(IDevPtr dev)
{
    ADBG("Alps::setInpInfo: sigio supported=" << xf86SIGIOSupported());
    InputInfoPtr    ii = _inpInfo;
    ii->name                    = dev->identifier;
    ii->type_name               = (char*)XI_MOUSE;
    ii->device_control          = &sDeviceControl;
    ii->read_input              = &sReadInput;
    ii->control_proc            = sControlProc;
    ii->close_proc              = sCloseProc;
    ii->switch_mode             = sSwitchMode;
    ii->conversion_proc         = sConversionProc;
    ii->reverse_conversion_proc = 0;
    ii->dev                     = 0;
    ii->private_flags           = 0;
    ii->priv8                   = this;
    ii->flags                   = XI86_POINTER_CAPABLE |
                                  XI86_SEND_DRAG_EVENTS;
    ii->conf_idev               = dev;
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
    ii->motion_history_proc     = xf86GetMotionEvents;
    ii->history_size            = 0;
#endif
    ii->always_core_feedback    = 0;
} // Alps::setInpInfo


////////////////////////////////////////////////////////////////////////
Alps::Alps(inpinfo_t*  inpinfo) :
    _inpInfo(inpinfo)
{
} // Alps::Alps


////////////////////////////////////////////////////////////////////////
void Alps::configure(const Config& conf)
{
    delete _buttonHandler;
    delete _padHandler;
    _midEmulationDelayMilliSec = conf.delay_ms;
    _buttonHandler = new button::Handler(conf.delay_ms != 0);
    _padHandler = new pad::Handler();
    for (unsigned di = 0;  di != 2;  ++di)
    {
        const Config::Dim&  d = conf.dim[di];
        _padHandler->set(di,  d.min, d.max,  d.nc, d.dc,  d.ne, d.de,
                         d.jump);
    }
} // Alps::configure


#if !defined(NO_SHM)
////////////////////////////////////////////////////////////////////////
void Alps::reconf()
{
    Config conf;

    conf.shm_get(*_shm);
    _shm->mailBox()->cmd = 0;
    configure(conf);
} // Alps::reconf
#endif


////////////////////////////////////////////////////////////////////////
Bool Alps::deviceControl(DeviceIntPtr dev, int mode)
{
    Bool ret;

    ADBG("deviceControl dev=" << dev << ", mode=" << mode <<
         ", ->dev=" << (_inpInfo ? _inpInfo->dev : 0));
    switch (mode)
    {
     case DEVICE_INIT:
        ret = devInit(dev);
        break;
     case DEVICE_ON:
        ret = devOn(dev);
        break;
     case DEVICE_OFF:
        ret = devOff(dev);
        break;
     case DEVICE_CLOSE:
        ret = devClose(dev);
        break;
     default:
        ret = BadValue;
    }
    ADBG("deviceControl ret=" << ret);
    return ret;
} // Alps::deviceControl


////////////////////////////////////////////////////////////////////////
void Alps::readInput()
{
    ADBG(strnow() << " Alps::readInput");
    _timer2BeCalled = 0;
    // int sigstate = xf86BlockSIGIO();
    while (readEvent())
    {
        if (_timer2BeCalled && (_timer2BeCalled->atMillis < GetTimeInMillis()))
        {
            processTimer(_timer2BeCalled->lrButton, false);
            _timer2BeCalled = 0;
        }
    }
    // xf86UnblockSIGIO(sigstate);
    if (_timer2BeCalled)
    {
        ADBG("TimerSet: atMillis=" << _timer2BeCalled->atMillis);
        // int sigstate = xf86BlockSIGIO();
        _timer = TimerSet(_timer, TimerAbsolute, _timer2BeCalled->atMillis,
                          &timerFunc, _timer2BeCalled);
        // xf86UnblockSIGIO(sigstate);

    }
    ADBG(strnow() << " Alps::readInput Done");
} // Alps::readInput


////////////////////////////////////////////////////////////////////////
bool Alps::readEvent()
{
    ADBG("Alps::readEvent");
    struct input_event  ev;
    ucp_t               pb0 = (unsigned char*)(&ev);
    ucp_t               pb = pb0,  pbEnd = pb0 + sizeof(ev);
    while (pb < pbEnd)
    {
        int     c = XisbRead(_xisBuffer);
        uc_t    u = (uc_t)(c);
        *pb = u;
        if (c < 0)
        {
            ADBG("XisbRead failed c=" << c << ", read=" << pb-pb0 <<
                 " |ev|=" << sizeof(ev));
            pb = pbEnd+1; // make: gotEvent=false
        }
        else
        {
            ++pb;
        }
    }
    bool    gotEvent = (pb == pbEnd);
    ADBG("gotEvent=" << int(gotEvent));
    if (gotEvent)
    {
        ADBG(strnow() << hex <<
             " type=0x" << ev.type << ", code=0x" << ev.code <<
             ", v=" << ev.value << dec);
        processEvent(ev);
    }
    return gotEvent;
} // Alps::readEvent


////////////////////////////////////////////////////////////////////////
void Alps::processEvent(const struct input_event& ev)
{
    ADBG("Alps::processEvent");
#if !defined(NO_SHM)
    if (_shm->ok() && _shm->mailBox()->cmd)
    {
        reconf();
    }
#endif
    switch (ev.type)
    {
     case EV_KEY:
        keyEvent(ev);
        break;
     case EV_ABS:
        absEvent(ev);
        break;
     case EV_SYN:
        absSyn();
     default: ;
    }
} // Alps::processEvent


////////////////////////////////////////////////////////////////////////
void Alps::keyEvent(const struct input_event& ev)
{
    ADBG("Alps::keyEvent");
    button::LogicalEvent  logical;
    switch (ev.code)
    {
     case BTN_LEFT:
        logical.button = button::ButtonLeft;
        break;
     case BTN_RIGHT:
        logical.button = button::ButtonRight;
        break;
     default:;
    }
    if (logical.button != button::NoButton)
    {
        _buttonHandler->physical2logical(logical, logical.button, ev.value);
        if (logical.action == button::LogicalEvent::Timer)
        {
            ADBG("Sending Timer, delay_ms=" << _midEmulationDelayMilliSec);
            _timer2BeCalled = &_timerCallParam[logical.button];
            _timer2BeCalled->atMillis =
                GetTimeInMillis() + _midEmulationDelayMilliSec;
        }
        else
        {
            postButtonEvent(logical);
        }
    }
} // Alps::keyEvent


////////////////////////////////////////////////////////////////////////
Alps::ul_t Alps::timerFunc(OsTimerPtr timer, ul_t /* now */, void* vp)
{
    TimerCallParam*  alpslr = (TimerCallParam*)(vp);
    return  alpslr->p->processTimer(alpslr->lrButton, true);
} // Alps::timerFunc


////////////////////////////////////////////////////////////////////////
Alps::ul_t  Alps::processTimer(unsigned lrButton, bool block)
{
    ADBG(strnow() << " Alps::processTimer lrButton=" << lrButton <<
         ", block=" << block);
    button::LogicalEvent  logical;
    _buttonHandler->timer(logical, lrButton);
    if (logical.action != button::LogicalEvent::Idle)
    {
        if (block)
        {
            int  sigstate = xf86BlockSIGIO();
            postButtonEvent(logical);
            xf86UnblockSIGIO(sigstate);
        }
        else
        {
            postButtonEvent(logical);
        }
    }
    return 0;
} // Alps::processTimer


////////////////////////////////////////////////////////////////////////
void  Alps::postButtonEvent(const button::LogicalEvent&  logical)
{
    int  buttonId = _bhe2xbi[logical.button];
    ADBG("postButtonEvent: buttonId=" << buttonId << ", action=" <<
         int(logical.action));
    switch (logical.action)
    {
     case button::LogicalEvent::Down:
        xf86PostButtonEvent(_inpInfo->dev, FALSE, buttonId, 1, 0, 0);
        break;
     case button::LogicalEvent::Up:
        xf86PostButtonEvent(_inpInfo->dev, FALSE, buttonId, 0, 0, 0);
        break;
     case button::LogicalEvent::DownUp:
        xf86PostButtonEvent(_inpInfo->dev, FALSE, buttonId, 1, 0, 0);
        xf86PostButtonEvent(_inpInfo->dev, FALSE, buttonId, 0, 0, 0);
        break;
     default:
        ADBG("Bad logical.state");
    }
} // Alps::postButtonEvent


////////////////////////////////////////////////////////////////////////
void Alps::absEvent(const struct input_event& ev)
{
    ADBG("Alps::absEvent: code=" << ev.code << ", v=" << ev.value);
    switch (ev.code)
    {
     case ABS_X:
        _absX = ev.value;
        break;
     case ABS_Y:
        _absY = ev.value;
        break;
     default:;
    }
} // Alps::absEvent


////////////////////////////////////////////////////////////////////////
void Alps::absSyn()
{
    ADBG("Alps::absSyn");
    int  dsx, dsy;
    _padHandler->getDeltaScreen(dsx, dsy, _absX, _absY);
    if ((dsx != 0) || (dsy != 0))
    {
        ADBG("dsx=" << dsx << ", dsy=" << dsy);
        xf86PostMotionEvent(_inpInfo->dev, 0, 0, 2, dsx, dsy);
    }
} // Alps::absSyn


////////////////////////////////////////////////////////////////////////
int  Alps::controlProc(xDeviceCtl *control)
{
    ADBG("Alps::controlProc");
    return Success;
} // Alps::controlProc


////////////////////////////////////////////////////////////////////////
void  Alps::closeProc()
{
    ADBG("Alps::closeProc");
} // Alps::closeProc


////////////////////////////////////////////////////////////////////////
int  Alps::switchMode(int mode)
{
    ADBG("Alps::switchMode");
    return Success;
} // Alps::switchMode


////////////////////////////////////////////////////////////////////////
Bool  Alps::conversionProc
(
    int     first,
    int     num,
    int     v0,
    int     v1,
    int&    x,
    int&    y
)
{
    ADBG("Alps::conversionProc");
    Bool  rc = FALSE;
    if (first == 0 && num == 2)
    {
	rc = TRUE;
        x = v0;
        y = v1;
    }
    return rc;
} // Alps::conversionProc


////////////////////////////////////////////////////////////////////////
static void  alpsControl(DeviceIntPtr, PtrCtrl*)
{
    ADBG("alpsControl");
} // alpsControl


////////////////////////////////////////////////////////////////////////
Bool  Alps::devInit(DeviceIntPtr dev)
{
    ADBG("Alps::devInit");
    uc_t    buttons_map[MaxButtons + 1];
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
    Atom    btn_labels[MaxButtons] = { 0 };
    Atom    axes_labels[2] = { 0 };

    InitAxesLabels(axes_labels, 2);
    InitButtonLabels(btn_labels, MaxButtons);
#endif

    for (unsigned  i = 0;  i <= MaxButtons;  buttons_map[i] = i, ++i);

    dev->publik.on = FALSE;

    InitPointerDeviceStruct((DevicePtr)(dev), buttons_map,
                            MaxButtons,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                            btn_labels,
#endif
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
			    miPointerGetMotionEvents,
#elif GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 3
			    GetMotionHistory,
#endif

                            &alpsControl,

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
			    miPointerGetMotionBufferSize()
#else
			    GetMotionHistorySize(), 2
#endif
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                            , axes_labels
#endif
			    );


    /* X valuator */
    xf86InitValuatorAxisStruct(dev, 0, 0, -1, 1, 0, 1);
    xf86InitValuatorDefaults(dev, 0);
    /* Y valuator */
    xf86InitValuatorAxisStruct(dev, 1, 0, -1, 1, 0, 1);
    xf86InitValuatorDefaults(dev, 1);

    xf86MotionHistoryAllocate(_inpInfo);

    return Success;
} // Alps::devInit


#if !defined(EVIOCGRAB)
#define EVIOCGRAB  _IOW('E', 0x90, int) /* Grab/Release device */
#define EV_SYN			0x00
#endif

////////////////////////////////////////////////////////////////////////
Bool  Alps::devOn(DeviceIntPtr dev)
{
    InputInfoPtr    ii = _inpInfo;
    ADBG("Alps::devOn fd=" << ii->fd);
    openSerial();
    if (ok())
    {
        xf86FlushInput(ii->fd);
        int ret = eventIoctl(ii->fd, EVIOCGRAB, (void*)1);
        ADBG("ret=" << ret);
        if (ret < 0)
        {
            ErrorF("Failed to grab event device ret=%d", ret);
            setError(eGrabEvent);
        }
    }
    if (ok() && devReset() && newBuffer(0x40))
    {
        xf86AddEnabledDevice(ii);
        dev->publik.on = TRUE;
    }
    return (ok() ? Success : !Success);
} // Alps::devOn


////////////////////////////////////////////////////////////////////////
Bool  Alps::devOff(DeviceIntPtr dev)
{
    ADBG("Alps::devOff fd=" << _inpInfo->fd);
    // Folloing synaptics's DeviceOff logic
    if (_inpInfo->fd != -1)
    {
        xf86RemoveEnabledDevice(_inpInfo);
        release();
    }
    dev->publik.on = FALSE;
    return Success;
} // Alps::devOff


////////////////////////////////////////////////////////////////////////
Bool  Alps::devClose(DeviceIntPtr dev)
{
    ADBG("Alps::devClose");
    Bool  ret = devOff(dev);
    return ret;
} // Alps::devClose


////////////////////////////////////////////////////////////////////////
static inline bool  test_bit(unsigned bit, unsigned long* longbits)
{
    unsigned  li  = bit / bitsPerLong;
    unsigned  obi = bit % bitsPerLong;
    bool      on = (((1ul << obi) & longbits[li]) != 0);
    return on;
} // test_bit


////////////////////////////////////////////////////////////////////////
bool Alps::isTouchPad(int  fd)
{
    bool           isTP = true;
    unsigned long  evbits[nULongs], *evbp = &evbits[0];
    ADBG("Alps::isTouchPad");

    isTP = isTP && (eventIoctl(fd, EVIOCGBIT(0, EV_MAX), evbp) >= 0);
    isTP = isTP &&
           test_bit(EV_SYN, evbp) &&
           test_bit(EV_ABS, evbp) &&
           test_bit(EV_KEY, evbp);

    isTP = isTP && (eventIoctl(fd, EVIOCGBIT(EV_ABS, KEY_MAX), evbp) >= 0);
    isTP = isTP &&
           test_bit(ABS_X, evbp) &&
           test_bit(ABS_Y, evbp) &&
           test_bit(ABS_PRESSURE, evbp);

    isTP = isTP && (eventIoctl(fd, EVIOCGBIT(EV_KEY, KEY_MAX), evbp) >= 0);
    isTP = isTP && test_bit(BTN_TOOL_FINGER, evbp);

    ADBG("Alps::isTouchPad isTP=" << int(isTP));
    return isTP;
} // Alps::isTouchPad



////////////////////////////////////////////////////////////////////////
int Alps::eventIoctl(int fd, int cmd, void* vp)
{
    ADBG("eventIoctl: fd=" << fd << ", cmd=0x" << hex << cmd <<
         ", vp=" << vp << dec);
    int  ret = 0;
    while (((ret = ioctl(fd, cmd, vp)) != -1) && (errno == EINTR));
    ADBG("eventIoctl: ret=" << ret);
    return ret;
} // Alps::eventIoctl


////////////////////////////////////////////////////////////////////////
bool  Alps::openSerial()
{
    ADBG("Alps::openSerial");
    setDevice();
    if (ok())
    {
        InputInfoPtr    ii = _inpInfo;
        ii->fd = xf86OpenSerial(ii->options);
        ADBG("fd=" << ii->fd);
        if (ii->fd == -1)
        {
            xf86Msg(X_WARNING, "%s: cannot open input device\n", ii->name);
            setError(eOpenDevice);
        }
        else
        {
            xf86ErrorFVerb(6, "port opened successfully\n");
        }
    }
    return ok();
} // Alps::openSerial


////////////////////////////////////////////////////////////////////////
void  Alps::closeSerial()
{
    ADBG("Alps::closeSerial");
    if (_inpInfo)
    {
        if (_inpInfo->fd != -1)
        {
             ADBG("Closing fd=" << _inpInfo->fd);
             xf86CloseSerial(_inpInfo->fd);
             _inpInfo->fd = -1;
        }
    }
} // Alps::closeSerial


////////////////////////////////////////////////////////////////////////
bool  Alps::setDevice()
{
    ADBG("Alps::setDevice");
    bool  found = false;
    static const char*  devEventX =    "/dev/input/eventX";
    char                fname[32];  //  0123456789 123456
    unsigned            Xi = strlen(devEventX) - 1;
    strcpy(fname, devEventX);
    for (unsigned  n = 0;  !found && (n <= 9);  ++n)
    {
        fname[Xi] = "0123456789"[n];
        int  fd = open(fname, O_RDONLY);
        if (fd >= 0)
        {
            ADBG("fd=" << fd << "  fname=" << fname);
            found = isTouchPad(fd);
            close(fd);
        }
    }
    ADBG("found=" << found);
    if (found)
    {
        xf86Msg(X_PROBED, "alps: set device to %s\n", fname);
        xf86ReplaceStrOption(_inpInfo->options, "Device", fname);
    }
    else
    {
        ErrorF("alps: Not a touch pad\n");
        setError(eNoTouchPad);
    }
    return ok();
} // Alps::setDevice


////////////////////////////////////////////////////////////////////////
bool  Alps::newBuffer(unsigned size)
{
    ADBG("Alps::newBuffer  size=" << size);
    _xisBuffer = XisbNew(_inpInfo->fd, size);
    if (_xisBuffer == 0)
    {
       ErrorF("alps: XisbNew failed\n");
       setError(eXISBuffer);
    }
    return ok();
} // Alps::newBuffer


////////////////////////////////////////////////////////////////////////
void  Alps::delBuffer()
{
    ADBG("Alps::delBuffer _xisBuffer=" << _xisBuffer);
    if (_xisBuffer)
    {
        XisbFree(_xisBuffer);
        _xisBuffer = 0;
    }
} // Alps::delBuffer


////////////////////////////////////////////////////////////////////////
bool  Alps::devReset()
{
#if 0
    static const uc_t  seq[] =
    {
        PS2_CMD_SET_DEFAULT,
        PS2_CMD_SET_SCALING_2_1,
        PS2_CMD_SET_SCALING_2_1,
        PS2_CMD_SET_SCALING_2_1,
        PS2_CMD_DISABLE,

        PS2_CMD_DISABLE,
        PS2_CMD_DISABLE,
        PS2_CMD_DISABLE,
        PS2_CMD_DISABLE,
        PS2_CMD_ENABLE,

        PS2_CMD_SET_SCALING_1_1,
        PS2_CMD_SET_SCALING_1_1,
        PS2_CMD_SET_SCALING_1_1,
        PS2_CMD_DISABLE,

        PS2_CMD_DISABLE,
        PS2_CMD_DISABLE,
        PS2_CMD_DISABLE,
        PS2_CMD_DISABLE,
        PS2_CMD_ENABLE,
    };

    ADBG("Alps::devReset");
    int            fd = _inpInfo->fd;

    xf86FlushInput(fd);
    for (cucp_t  pb = &seq[0],  pbEnd = pb + sizeof(seq)/sizeof(seq[0]);
         ok() && (pb != pbEnd);  ++pb)
    {
        putByteAck(*pb);
    }
    // Wait for the channel to go silent, which means we're in sync
    if (ok())
    {
        uc_t  buffer[64];
        while (xf86WaitForInput(fd, 250000) > 0)
        {
	    xf86ReadSerial(fd, &buffer[0], 64);
        }
    }
#endif
    return ok();
} // Alps::devReset


////////////////////////////////////////////////////////////////////////
bool  Alps::putByteAck(uc_t b)
{
    ADBG("Alps::putByteAck b=0x" << hex << setw(2) << (unsigned)(b) << dec);
    int     fd = _inpInfo->fd;
    uc_t    ack;

    int  wrc = xf86WriteSerial(fd, &b, 1);
    ADBG("wrc=" << wrc);
    if (wrc != 1)
    {
        ErrorF("alps: error xf86WriteSerial\n");
        setError(eSerialPut);
    }
    if (ok() && getByte(ack))
    {
        if (ack != PS2_ACK)
        {
            ErrorF("alps: wrong acknowledge  0x%02x\n", ack);
            setError(eNoAck);
        }
    }
    return ok();
} // Alps::putByteAck


////////////////////////////////////////////////////////////////////////
bool Alps::getByte(uc_t& b)
{
    ADBG("Alps::getByte");
    int            fd = _inpInfo->fd;
    if (xf86WaitForInput(fd, 50000) > 0)
    {
        if (xf86ReadSerial(fd, &b, 1) != 1)
        {
            ErrorF("alps: No byte read\n");
            setError(eSerialGet);
        }
    }
    else
    {
        ErrorF("alps: timeout xf86WaitForInput\n");
        setError(eSerialTimeout);
    }
    return ok();
} // Alps::getByte


////////////////////////////////////////////////////////////////////////
void Alps::setErrorFnLn(int e, const char* fn, unsigned ln)
{
    if (_error == eOK)
    {
        _error = e;
        ADBG("setError('" << fn << "'[" << ln << "] e=" << e);
    }
} // Alps::setErrorFnLn
