// -*- c++ -*-
#if !defined(Alps_H)
#define Alps_H

#define class _klAss
#define bool  _bOol
#define private  priv8
#define public   publik
extern "C"
{
#   include <X11/extensions/XIproto.h>
#   include <input.h>
#   include <xf86str.h>
}
#undef class
#undef bool
#undef private
#undef public
#undef xalloc

namespace button { class  Handler; class LogicalEvent; }
namespace pad { class  Handler;  }
namespace shm_config { class Handler; }

class Config;


class Alps
{
 public:

    typedef unsigned char           uc_t;
    typedef unsigned long           ul_t;
    typedef uc_t*                   ucp_t;
    typedef uc_t const*             cucp_t;
    typedef struct _InputDriverRec  driver_t;
    typedef struct _LocalDeviceRec  inpinfo_t;
    typedef struct _OsTimerRec*     OsTimerPtr;

    enum
    {
        eOK,
        eAllocInp,
        eOpenDevice,
        eXISBuffer,
        eGrabEvent,
        eSerialPut,
        eSerialGet,
        eSerialTimeout,
        eNoAck,
        eNoTouchPad
    };

    static const char*  version;
    static void*  setup(void* module, void* options, int* errmaj, int* errmin);

    virtual ~Alps();
    inpinfo_t*          inpinfo() const { return _inpInfo; }
    bool                ok() const { return _error == 0; }
    int                 error() const { return _error; }

 private:

    enum
    {
        MaxButtons = 3
    };

    Alps(inpinfo_t*);
    static void         identify(int flags);
    static inpinfo_t*   preinit(driver_t* drv, IDevPtr dev, int flags);
    static void         uninit(driver_t* drv, inpinfo_t* inpInfo, int flags);
    static Alps*        thisAlps(DeviceIntPtr device);
    static Alps*        thisAlps(inpinfo_t* inpInfo);

    static Bool         sDeviceControl(DeviceIntPtr dev, int what)
                            { return thisAlps(dev)->deviceControl(dev, what); }
    static void         sReadInput(inpinfo_t* ii)
                            { thisAlps(ii)->readInput(); }
    static int          sControlProc(inpinfo_t *ii, xDeviceCtl* control)
                            { return thisAlps(ii)->controlProc(control); }
    static void         sCloseProc(inpinfo_t* ii)
                            { thisAlps(ii)->closeProc(); }
    static int          sSwitchMode(
                            ClientPtr client, DeviceIntPtr dev, int mode)
                            { return thisAlps(dev)->switchMode(mode); }
    static Bool         sConversionProc(
                           inpinfo_t* ii,
                           int first, int num, int v0,
                           int v1, int v2, int v3, int v4,
                           int v5, int* x, int* y)
                           { return thisAlps(ii)->conversionProc(
                               first,  num,  v0, v1,  *x,  *y);
                           }

    static ul_t         timerFunc(OsTimerPtr timer, ul_t now, void* vp);


    // For the member methods, trying to keep order of top-down call level
    void                configure(const Config&);
    void                reconf();

    Bool                deviceControl(DeviceIntPtr dev, int what);
    void                readInput();
    bool                readEvent();
    void                processEvent(const struct input_event&);

    void                keyEvent(const struct input_event&);
    void                padEvent(const struct input_event&);
    ul_t                processTimer(unsigned lrButton, bool block);
    void                postButtonEvent(const button::LogicalEvent&);

    void                absEvent(const struct input_event&);
    void                absSyn();

    int                 controlProc(xDeviceCtl *control);
    void                closeProc();
    int                 switchMode(int mode);
    Bool                conversionProc(
                            int first, int num, int v0, int v1, int& x, int& y);


    void                setInpInfo(IDevPtr dev);
    void                release();
    void                setErrorFnLn(int e, const char* fn, unsigned ln);
#define setError(e)  setErrorFnLn(e, __FILE__, __LINE__)

    Bool                devInit(DeviceIntPtr dev);
    Bool                devOn(DeviceIntPtr dev);
    Bool                devOff(DeviceIntPtr dev);
    Bool                devClose(DeviceIntPtr dev);

    bool                openSerial();
    void                closeSerial();
    bool                setDevice();
    bool                isTouchPad(int fd);
    int                 eventIoctl(int fd, int cmd, void*);
    bool                newBuffer(unsigned size);
    void                delBuffer();

    bool                devReset();
    bool                putByteAck(uc_t b);
    bool                getByte(uc_t& b);

    // Not publicly available on purpose!
    Alps(driver_t* drv, IDevPtr dev, int flags);
    Alps(const Alps&);
    Alps& operator=(const Alps&);

    struct TimerCallParam
    {
         TimerCallParam(Alps* pa=0, unsigned lr=0, ul_t at=0) :
             p(pa), lrButton(lr), atMillis(at) {}
         Alps*      p;
         unsigned   lrButton;
         ul_t       atMillis;
    };

    static const int    _bhe2xbi[3];

    inpinfo_t*              _inpInfo;
    int                     _error;
    struct _XISBuffer*      _xisBuffer;

    button::Handler*        _buttonHandler;
    OsTimerPtr              _timer;
    TimerCallParam          _timerCallParam[2];
    TimerCallParam*         _timer2BeCalled;

    pad::Handler*           _padHandler;
    unsigned                _midEmulationDelayMilliSec; 
    unsigned                _absX;
    unsigned                _absY;

    shm_config::Handler*    _shm;
}; // Alps

#endif /* Alps_H */
