// GPL
// Author:  Yotam Medini  yotam.medini@gmail.com -- Created: 2006/March/17
#include <fstream>
#include <iomanip>
#include "ButtonHandler.h"
#include "dbg.h"

namespace button 
{

////////////////////////////////////////////////////////////////////////
Handler::Handler(bool  emulateMid) :
    _emulateMid(emulateMid)
{
    ADBG("emulateMid=" << int(emulateMid));
    _lrButton[ButtonLeft]  = LRButton(false, ButtonLeft);
    _lrButton[ButtonRight] = LRButton(false, ButtonRight);
    _midButtonDown = false;
} // Handler::Handler


////////////////////////////////////////////////////////////////////////
Handler::~Handler()
{
} // Handler::Handler


////////////////////////////////////////////////////////////////////////
void
Handler::physical2logical(LogicalEvent& logical, unsigned lr_button, bool down)
{
    ADBG("lr_button=" << lr_button << ", down=" << down);
    unsigned    lr_other = 1 - lr_button;
    LRButton&   currButton = _lrButton[lr_button];
    LRButton&   other      = _lrButton[lr_other];
    currButton.physicalDown = down;
    logical.button = lr_button; // may change
    if (emulationOn())
    {
        ADBG("Emulation On");
        if (down)
        {
            if (_midButtonDown)
            {
                logical.button = NoButton;
                logical.action = LogicalEvent::Idle;
            }
            else
            {
                if (other.logicalState == LogicalDownPending)
                {
                    other.logicalState = LogicalUp;
                    _midButtonDown = true;
                    logical.button = ButtonMid;
                    logical.action = LogicalEvent::Down;
                }
                else
                {
                    currButton.logicalState = LogicalDownPending;
                    logical.action = LogicalEvent::Timer;
                }
            }
        }
        else // up
        {
            // Initial values that change here
            logical.action = LogicalEvent::Up; 

            switch (currButton.logicalState)
            {
             case LogicalUp:
                // _midButtonDown must be true!
                if (other.physicalDown)
                {
                    logical.action = LogicalEvent::Idle;
                }
                else
                {
                    logical.button = ButtonMid;
                    _midButtonDown = false;
                }
                break;
             case LogicalDown:
                logical.action = LogicalEvent::Up;
                break;
             case LogicalDownPending:
                logical.action = LogicalEvent::DownUp;
                break;
            }
            currButton.logicalState = LogicalUp;
        }
    }
    ADBG("logical: button=" << logical.button << 
         " action=" << int(logical.action));
} // Handler::physical2logical


////////////////////////////////////////////////////////////////////////
void  Handler::timer(LogicalEvent& logical, unsigned lr)
{
    logical.button = lr;
    LRButton&  currButton = _lrButton[lr];
    ADBG("timer: lr=" << lr << " logState=" << int(currButton.logicalState));
    if (currButton.logicalState == LogicalDownPending)
    {
        logical.action = LogicalEvent::Down;
        currButton.logicalState = LogicalDown;
    }
    else
    {
        logical.action = LogicalEvent::Idle;
    }
    ADBG("timer/logical: butt=" << logical.button << 
         " action=" << int(logical.action));
} // Handler::timer


} // namespace button
