// -*- c++ -*-
// GPL
// Author:  Yotam Medini  yotam.medini@gmail.com -- Created: 2006/March/17
#if !defined(ButtonHandler_H)
#define ButtonHandler_H

#include <os.h>


namespace button
{


enum 
{ 
    ButtonTimer = -2, 
    NoButton    = -1, 
    ButtonLeft  =  0, 
    ButtonRight =  1, 
    ButtonMid   =  2 
};


class LogicalEvent;


class Handler
{
 public:
    typedef unsigned long  ul_t;
    typedef ul_t           et_t; // Event-Time
    // typedef LogicalEvent   Logical;

    Handler(bool emulateMid = true);
    virtual ~Handler();

    void     physical2logical(LogicalEvent&, unsigned  lr_button, bool down);
    void     timer(LogicalEvent&, unsigned lr);
    bool     emulationOn() const { return _emulateMid; }

 private:

    enum LogicalState { LogicalUp, LogicalDownPending, LogicalDown};

    Handler(const Handler&);
    Handler& operator=(const Handler&);

    struct LRButton
    {
        LRButton(bool down=false, unsigned lr=0) : 
            physicalDown(false), logicalState(LogicalUp) {}
        bool          physicalDown;
        LogicalState  logicalState;
    };
    bool             _emulateMid;
    LRButton         _lrButton[2];
    bool             _midButtonDown;
}; // Handler


class LogicalEvent
{
 public:
    enum Action { Idle, Down, Up, DownUp, Timer};
    LogicalEvent(int b = NoButton, Action a = Idle) : 
        button(b), action(a) {}
    int      button;  // If NoButton, 
    Action   action;  // - then meaningless        
}; // LogicalEvent


} // namespace button

#endif /* ButtonHandler_H */
