// GPL
// Author:  Yotam Medini  yotam.medini@gmail.com -- Created: 2006/March/18
#include "PadHandler.h"
#include <cstdlib>
#include <iostream>
#include "dbg.h"

using namespace std;

namespace pad
{

////////////////////////////////////////////////////////////////////////
Range::Range(int b, int e, unsigned nc, unsigned dc, unsigned ne, unsigned de) :
    _begin(b),
    _end(e),
    _nc(nc),
    _dc(dc),
    _ne(ne),
    _de(de),
    m((b+e)/2), A(0), B(1), D(1)
{
    preCompute();
} // Range::Range


////////////////////////////////////////////////////////////////////////
void  Range::configure
(
     int        b, 
     int        e,
     unsigned   nc, 
     unsigned   dc, 
     unsigned   ne, 
     unsigned   de
)
{
    _begin = b;
    _end = e;
    _nc = nc; 
    _dc = dc;
    _ne = ne;
    _de = de; 
    preCompute();
} // Range::configure


////////////////////////////////////////////////////////////////////////
static unsigned long  euclid_gcd(long m, long n)
{
    if (m < 0) {m = -m;}
    if (n < 0) {n = -n;}
    if (m < n)
    {
        unsigned long  t = m;  m = n;  n = t;
    }
    while (n > 0)
    {
        unsigned long  r = m % n;
        m = n;
        n = r;
    }
    return m;
} // euclid_gcd


////////////////////////////////////////////////////////////////////////
// See LaTeX document
void Range::preCompute()
{
    long  r = (_end - _begin + 1) / 2;
    long  tr2de = 3 * r*r * _de;
    m = (_begin + _end) / 2;
    A = _ne * _dc - _nc * _de;
    B = tr2de * _nc;
    D = tr2de * _dc;
    long  gcd = euclid_gcd(euclid_gcd(A, B), D);
    A /= gcd;
    B /= gcd;
    D /= gcd;
    ADBG("A=" << A << ", B=" << B << ", D=" << D);
} // Range::preCompute


////////////////////////////////////////////////////////////////////////
long  Range::pad2screen(int p) const
{
    long    dx = long(p) - long(m);
    long    s = ((A * dx * dx + B) * dx) / D;
    return  s;
} // Range::pad2screen


////////////////////////////////////////////////////////////////////////
History::History(const Range& range, long s, long p, long d) :
    padPrevious(p), 
    screenStartRef(range.pad2screen(s)),
    screenCurrDelta(d)
{
    ADBG("History::History");
} // History::History


////////////////////////////////////////////////////////////////////////
int DimensionMoveManager::hmove(History& h, int vNew) const
{
    ADBG("hmove: vNew=" << vNew << 
         " padPrevious=" << h.padPrevious <<
         ", screenStartRef=" << h.screenStartRef << 
         ", screenCurrDelta=" << h.screenCurrDelta);
    int  delta = 0;
    int  dv = vNew - h.padPrevious;
    if (abs(dv) > _jumpSize)
    {
        h = History(_range, vNew, vNew, 0);
    }
    else
    {
        long    newScreenRef = _range.pad2screen(vNew);
        long    screenMoveDelta = newScreenRef - h.screenStartRef;
        delta = screenMoveDelta - h.screenCurrDelta;
        h.padPrevious = vNew;
        h.screenCurrDelta = screenMoveDelta;
    }    
    ADBG("DimensionMoveManager::hmove  delta=" << delta);
    return  delta;
} // DimensionMoveManager::hmove


////////////////////////////////////////////////////////////////////////
void  Handler::set
(
    unsigned    xy,
    int         padMin, 
    int         padMax,
    unsigned    nc, 
    unsigned    dc,
    unsigned    ne, 
    unsigned    de,
    unsigned    jumpSize
)
{
    ADBG("Handler::set " << xy << ' ' << padMin << ' ' << padMax << ' ' <<
         nc << ' ' << dc << ' ' << ne << ' ' << de << ' ' << jumpSize);
    _moveMgr[xy] = DimensionMoveManager(Range(padMin, padMax, nc, dc, ne, de),
                                        jumpSize);
} // Handler::setRect


////////////////////////////////////////////////////////////////////////
void  Handler::getDeltaScreen(int& dsx, int& dsy, int px, int py)
{
    ADBG("Handler::getDeltaScreen pad-coords: " << px << ' ' << py);
    dsx = _moveMgr[X].move(px);
    dsy = _moveMgr[Y].move(py);
} // Handler::getDeltaScreen


} // namespace
