// -*- c++ -*-
// GPL
// Author:  Yotam Medini  yotam.medini@gmail.com -- Created: 2006/March/17
#if !defined(PadHandler_H)
#define PadHandler_H

namespace pad
{

// typedef unsigned int   ui_t;
// typedef unsigned long  ul_t;
enum {X, Y, NXY};  enum {Min, Max};


// See LaTeX document
class Range
{
 public:
    Range(int m = 0, int M = 1,
          unsigned nc=1, unsigned dc=1, unsigned ne=1, unsigned de=1);
    void        configure(
                    int        b, 
                    int        e,
                    unsigned   nc, 
                    unsigned   dc, 
                    unsigned   ne, 
                    unsigned   de);
#if 0
    void        set(int m, int M) { _be[Min] = m;  _be[Max] = M; preCompute(); }
    void        setS2PRatios(unsigned nc, unsigned dc, unsigned ne, unsigned de);
#endif
    int         mid() const { return m; }
    long        pad2screen(int p) const;

 private:
    void        preCompute();
    // int       _be[2];  // [from, to)  STL-like
    int       _begin, _end;  // [from, to)  STL-like
    unsigned  _nc, _dc;
    unsigned  _ne, _de;
    long      m, A, B, D;
};  // Range


class History
{
 public:
    History(const Range& range = Range(), long s = 0, long p = 0, long d = 0);
    long    padPrevious;     // Previous pad position
    long    screenStartRef;  // Just reference for pixel delta, not absolute
    int     screenCurrDelta; // Current total screen delta of monotonic move
}; // History


class DimensionMoveManager
{
 public:
    DimensionMoveManager(const Range& range = Range(), unsigned jumpSize = 0x100) :
        _range(range),
        _history(_range, _range.mid(), _range.mid(), 0),
        _jumpSize(jumpSize)
        {}
    virtual ~DimensionMoveManager() {}
    int     move(int vNew) { return hmove(_history, vNew); }

 private:
    int         hmove(History& myHistory, int vNew) const;
    Range       _range;
    History     _history;
    int         _jumpSize; // Above it - movement is ignored.
}; // DimensionMoveManager


class Handler
{
 public:
    Handler() {}
    virtual ~Handler() {}

    void    set(unsigned xy,  // X or Y
                int padMin, int padMax,
                                                    // n?/d? Screen to Pad ratios
                unsigned  nc = 1, unsigned dc = 1,  // nc/dc : Center
                unsigned  ne = 1, unsigned de = 1,  // ne/de : Edge
                unsigned jumpSize = 0x100);
    void    getDeltaScreen(int& dsx, int& dsy, int px, int py);

 private:
    Handler(const Handler&);
    Handler& operator=(const Handler&);
    DimensionMoveManager  _moveMgr[NXY];
}; // Handler

} // namespace pad

#endif /* PadHandler_H */
