#if !defined(NO_SHM)

#include "SharedMemConfig.h"
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/ipc.h>

#if defined(SHM_TEST)
#   include <iostream>
#   include <iomanip>
#   include <cstdio>
#   include <cstdlib>
#   include <unistd.h>
#   define ErrMsg(fmt, ...) fprintf(stderr, fmt, __VA_ARGS__)
#else
#   define bool     _bOol
#   define class    _klAss
#   define new      _nEw
#   define private  priv8
    extern "C"
    {
#   include <xf86.h>
    }
#   undef bool
#   undef class
#   undef new
#   undef private
#   define ErrMsg(fmt, ...) xf86Msg(X_ERROR, "alps: " fmt, __VA_ARGS__)
#endif

namespace shm_config
{

////////////////////////////////////////////////////////////////////////
Handler::Handler() :
    _key(SHM_KEY),
    _mailBox(0)
{
     _id = shmget(_key, 0, 0);
     if (_id != -1)
     {
         ErrMsg("SHM_KEY=%d already exists\n", SHM_KEY);
         struct shmid_ds ds;
         int rc = shmctl(_id, IPC_STAT, &ds);
         ErrMsg("rc=%d, shm_nattch=%lu\n", rc, ds.shm_nattch);
         if (ds.shm_nattch > 0)
         {
             ErrMsg("SHM_KEY=%d, is in use\n", SHM_KEY);
             _key = IPC_PRIVATE;
         }
         else
         {
             ErrMsg("Remove SHM_KEY=%d\n", SHM_KEY);
             shmctl(_id, IPC_RMID, 0);
         }
     }
     _id = shmget(_key,  sizeof(MailBox), 0774 | IPC_CREAT);
     if (_id == -1)
     {
         ErrMsg("shmget(%d) failed\n", _key);
     }
     else
     {
         if (_key == IPC_PRIVATE)
         {
             ErrMsg("shmget(IPC_PRIVATE) returned id=%d\n", _id);
         }
         _mailBox = (MailBox*)shmat(_id, 0, 0);
         if ((long)(_mailBox) == -1L)
         {
             ErrMsg("shmat(%d) failed\n", _id);
             _mailBox = 0;
         }
     }
} // Handler::Handler


////////////////////////////////////////////////////////////////////////
Handler::~Handler()
{
    if (_mailBox)
    {
        shmdt(_mailBox);
    }
    if (_id != -1)
    {
        shmctl(_id, IPC_RMID, 0);
    }
} // Handler::~Handler

} // shm_config


#endif


#if defined(SHM_TEST)

// g++ -g -o /tmp/shmcfgt -DSHM_TEST=1 SharedMemConfig.cc

int main(int argc, char** argv)
{
    using namespace shm_config;

    int rc = 0;
    Handler*  shm_cfg = Handler::create();
    std::cout << "pid=" << getpid() << "\n";
    std::cout << "Handler: key=" << shm_cfg->key() <<
        ", id=" << shm_cfg->id() << ", ok=" << shm_cfg->ok() << "\n";
    if (!shm_cfg->ok())
    {
        rc = 1;
    }
    else
    {
        Handler::MailBox *mb = shm_cfg->mailBox();
        for (int i = 0; i < SHM_N_VALUES; i++)
        {
            mb->options[i] = i;
        }
        int dream = (argc > 1 ? strtol(argv[1], 0, 0) : 10);
        std::cout << "Check mailbox, while sleeping for " << dream << 
            " seconds\n";
        for (; dream > 0; dream--)
        {
            sleep(1);
            if (mb->cmd)
            {
                for (int i = 0; i < SHM_N_VALUES; i++)
                {
                    std::cout << "options[" << std::setw(2) << i << "] = " << 
                        std::setw(5) << mb->options[i] << "\n";
                }
                mb->cmd = 0;
            }
        }
    }
    delete shm_cfg;
    std::cout << "pid=" << getpid() << ", rc=" << rc << "\n";
    return rc;
}

#endif
